module DBus
  module Wrapper
    class ProcessWatch
      class NameResolutionFatalError < Exception; end
      class NameResolutionRecoverableError < Exception; end

      DBUS_ID_REGEXP = /\A(:1\.[0-9]+)\z/.freeze

      def initialize(bus, process_name_regexp = /.*/)
        @bus                  = bus
        @process_name_regexp  = process_name_regexp
        
        @known_id_to_name     = {}
        @known_name_to_id     = {}
        @pending              = {}
        @events               = { :resolved => [], :registered => [], :unregistered => [] }
      end


      def start!
        @bus.monitor_signal("org.freedesktop.DBus", nil, "org.freedesktop.DBus", "NameOwnerChanged") { |message| on_name_owner_changed_signal(message) }
        @bus.method_call("org.freedesktop.DBus", "/", "org.freedesktop.DBus", "ListNames") { |message| on_list_names_return(message) }
      end


      def on(event_name, &block)
        raise ArgumentError, "Invalid event name #{event_name}" unless @events.keys.include? event_name.to_sym
        raise ArgumentError, "You must pass a block" unless block_given?

        @events[event_name.to_sym] << block
      end


      def resolve_name_to_id(name)
        @known_name_to_id[name]
      end


      def resolve_id_to_name(id)
        @known_id_to_name[id]
      end


      private 

      def trigger(event_name, *args)
        @events[event_name].each do |callback|
          callback.call *args
        end
      end


      def register(name, id)
        @known_name_to_id[name] = id
        @known_id_to_name[id] = name

        trigger :registered, name
      end


      def unregister(name)
        @known_id_to_name.delete @known_name_to_id[name]
        @known_name_to_id.delete name

        trigger :unregistered, name
      end


      def is_monitored_process_name?(name)
        not name.match(DBUS_ID_REGEXP) and not name == "org.freedesktop.DBus" and name.match(@process_name_regexp)
      end


      def on_name_owner_changed_signal(message)
        if is_monitored_process_name?(message.params[0])
          if message.params[1] == '' and message.params[2] =~ DBUS_ID_REGEXP
            register message.params[0], message.params[2]

          elsif message.params[2] == '' and message.params[1] =~ DBUS_ID_REGEXP
            unregister message.params[0]
          end
        end
      end


      def on_list_names_return(message)
        raise NameResolutionFatalError, "Unable to query list of DBus names (#{message})" if message.is_a? DBus::Error

        message.params.first.each do |name|
          if is_monitored_process_name?(name)
            @pending[name] = true
          end
        end

        if @pending.size == 0 
          trigger :resolved  

        else
          @pending.keys.each do |name|
            @bus.method_call("org.freedesktop.DBus", "/", "org.freedesktop.DBus", "GetNameOwner", [[ DBus::Type::STRING, name]]) do |message| 
              if message.is_a? DBus::Error
                raise NameResolutionRecoverableError, "Unable to get DBus ID of of DBus process #{name} (#{message}) - most probably it has disappeared during query"
              else
                register name, message.params.first

                @pending.delete name
                
                trigger :resolved  if @pending.size == 0 
              end
            end
          end
        end
      end
    end
  end
end
